@@ -10,7 +10,7 @@ describe "gemcutter's dependency API" do
     G
 
     bundle :install, :artifice => "endpoint"
-    out.should include("Fetching gem metadata from #{source_uri}")
+    expect(out).to include("Fetching gem metadata from #{source_uri}")
     should_be_installed "rack 1.0.0"
   end
 
@@ -21,7 +21,7 @@ describe "gemcutter's dependency API" do
     G
 
     bundle :install, :artifice => "endpoint"
-    out.should include("Could not find gem ' sinatra")
+    expect(out).to include("Could not find gem ' sinatra")
   end
 
   it "should handle nested dependencies" do
@@ -31,7 +31,7 @@ describe "gemcutter's dependency API" do
     G
 
     bundle :install, :artifice => "endpoint"
-    out.should include("Fetching gem metadata from #{source_uri}/...")
+    expect(out).to include("Fetching gem metadata from #{source_uri}/...")
     should_be_installed(
       "rails 2.3.2",
       "actionpack 2.3.2",
@@ -59,7 +59,7 @@ describe "gemcutter's dependency API" do
     bundle :install, :artifice => "endpoint"
 
     bundle "install --deployment", :artifice => "endpoint"
-    out.should include("Fetching gem metadata from #{source_uri}")
+    expect(out).to include("Fetching gem metadata from #{source_uri}")
     should_be_installed "rack 1.0.0"
   end
 
@@ -99,6 +99,20 @@ describe "gemcutter's dependency API" do
     should_be_installed("rails 2.3.2")
   end
 
+  it "doesn't fail if you only have a git gem with no deps when using --deployment" do
+    build_git "foo"
+    gemfile <<-G
+      source "#{source_uri}"
+      gem 'foo', :git => "file:///#{lib_path('foo-1.0')}"
+    G
+
+    bundle "install", :artifice => "endpoint"
+    bundle "install --deployment", :artifice => "endpoint", :exitstatus => true
+
+    expect(exitstatus).to eq(0)
+    should_be_installed("foo 1.0")
+  end
+
   it "falls back when the API errors out" do
     simulate_platform mswin
 
@@ -108,7 +122,7 @@ describe "gemcutter's dependency API" do
     G
 
     bundle :install, :fakeweb => "windows"
-    out.should include("\nFetching full source index from #{source_uri}")
+    expect(out).to include("Fetching source index from #{source_uri}")
     should_be_installed "rcov 1.0.0"
   end
 
@@ -124,7 +138,7 @@ describe "gemcutter's dependency API" do
       gem "rails"
     G
     bundle :install, :artifice => "endpoint_fallback"
-    out.should include("\nFetching full source index from #{source_uri}")
+    expect(out).to include("Fetching source index from #{source_uri}")
 
     should_be_installed(
       "activesupport 2.3.2",
@@ -143,8 +157,42 @@ describe "gemcutter's dependency API" do
       gem "rack"
     G
 
-    bundle :install, :artifice => "endpoint_marshal_fail"
-    out.should include("\nFetching full source index from #{source_uri}")
+    bundle :install, :verbose => true, :artifice => "endpoint_marshal_fail"
+    expect(out).to include("could not fetch from the dependency API, trying the full index")
+    should_be_installed "rack 1.0.0"
+  end
+
+  it "handles host redirects" do
+    gemfile <<-G
+      source "#{source_uri}"
+      gem "rack"
+    G
+
+    bundle :install, :artifice => "endpoint_host_redirect"
+    should_be_installed "rack 1.0.0"
+  end
+
+  it "handles host redirects without Net::HTTP::Persistent" do
+    gemfile <<-G
+      source "#{source_uri}"
+      gem "rack"
+    G
+
+    FileUtils.mkdir_p lib_path
+    File.open(lib_path("disable_net_http_persistent.rb"), "w") do |h|
+      h.write <<-H
+        module Kernel
+          alias require_without_disabled_net_http require
+          def require(*args)
+            raise LoadError, 'simulated' if args.first == 'openssl' && !caller.grep(/vendored_persistent/).empty?
+            require_without_disabled_net_http(*args)
+          end
+        end
+      H
+    end
+
+    bundle :install, :artifice => "endpoint_host_redirect", :requires => [lib_path("disable_net_http_persistent.rb")]
+    expect(out).to_not match(/Too many redirects/)
     should_be_installed "rack 1.0.0"
   end
 
@@ -155,18 +203,31 @@ describe "gemcutter's dependency API" do
     G
 
     bundle :install, :artifice => "endpoint_redirect"
-    out.should match(/Too many redirects/)
+    expect(out).to match(/Too many redirects/)
   end
 
-  it "uses the modern index when --full-index is passed" do
-    gemfile <<-G
-      source "#{source_uri}"
-      gem "rack"
-    G
+  context "when --full-index is specified" do
+    it "should use the modern index for install" do
+      gemfile <<-G
+        source "#{source_uri}"
+        gem "rack"
+      G
 
-    bundle "install --full-index", :artifice => "endpoint"
-    out.should include("Fetching source index from #{source_uri}")
-    should_be_installed "rack 1.0.0"
+      bundle "install --full-index", :artifice => "endpoint"
+      expect(out).to include("Fetching source index from #{source_uri}")
+      should_be_installed "rack 1.0.0"
+    end
+
+    it "should use the modern index for update" do
+      gemfile <<-G
+        source "#{source_uri}"
+        gem "rack"
+      G
+
+      bundle "update --full-index", :artifice => "endpoint"
+      expect(out).to include("Fetching source index from #{source_uri}")
+      should_be_installed "rack 1.0.0"
+    end
   end
 
   it "fetches again when more dependencies are found in subsequent sources" do
@@ -203,11 +264,8 @@ describe "gemcutter's dependency API" do
 
     bundle :install, :artifice => "endpoint_extra"
 
-    output = <<OUTPUT
-Fetching gem metadata from http://localgemserver.test/..
-Fetching gem metadata from http://localgemserver.test/extra/.
-OUTPUT
-    out.should include(output)
+    expect(out).to include("Fetching gem metadata from http://localgemserver.test/..")
+    expect(out).to include("Fetching source index from http://localgemserver.test/extra")
   end
 
   it "does not fetch every specs if the index of gems is large when doing back deps" do
@@ -217,7 +275,7 @@ OUTPUT
       end
       build_gem "missing"
       # need to hit the limit
-      1.upto(Bundler::Source::Rubygems::FORCE_MODERN_INDEX_LIMIT) do |i|
+      1.upto(Bundler::Source::Rubygems::API_REQUEST_LIMIT) do |i|
         build_gem "gem#{i}"
       end
 
@@ -273,10 +331,10 @@ OUTPUT
     G
 
     bundle :install, :artifice => "endpoint"
-    out.should include("Fetching gem metadata from #{source_uri}")
+    expect(out).to include("Fetching gem metadata from #{source_uri}")
   end
 
-  it "should install when EndpointSpecification with a bin dir owned by root", :sudo => true do
+  it "should install when EndpointSpecification has a bin dir owned by root", :sudo => true do
     sudo "mkdir -p #{system_gem_path("bin")}"
     sudo "chown -R root #{system_gem_path("bin")}"
 
@@ -297,7 +355,7 @@ OUTPUT
     bundle "install --binstubs", :artifice => "endpoint"
 
     gembin "rackup"
-    out.should == "1.0.0"
+    expect(out).to eq("1.0.0")
   end
 
   it "installs the bins when using --path and uses autoclean" do
@@ -308,7 +366,7 @@ OUTPUT
 
     bundle "install --path vendor/bundle", :artifice => "endpoint"
 
-    vendored_gems("bin/rackup").should exist
+    expect(vendored_gems("bin/rackup")).to exist
   end
 
   it "installs the bins when using --path and uses bundle clean" do
@@ -319,7 +377,7 @@ OUTPUT
 
     bundle "install --path vendor/bundle --no-clean", :artifice => "endpoint"
 
-    vendored_gems("bin/rackup").should exist
+    expect(vendored_gems("bin/rackup")).to exist
   end
 
   it "prints post_install_messages" do
@@ -329,7 +387,7 @@ OUTPUT
     G
 
     bundle :install, :artifice => "endpoint"
-    out.should include("Post-install message from rack:")
+    expect(out).to include("Post-install message from rack:")
   end
 
   it "should display the post install message for a dependency" do
@@ -339,8 +397,8 @@ OUTPUT
     G
 
     bundle :install, :artifice => "endpoint"
-    out.should include("Post-install message from rack:")
-    out.should include("Rack's post install message")
+    expect(out).to include("Post-install message from rack:")
+    expect(out).to include("Rack's post install message")
   end
 
   context "when using basic authentication" do
@@ -361,7 +419,7 @@ OUTPUT
       G
 
       bundle :install, :artifice => "endpoint_basic_authentication"
-      out.should_not include("#{user}:#{password}")
+      expect(out).not_to include("#{user}:#{password}")
       should_be_installed "rack 1.0.0"
     end
 
@@ -372,7 +430,7 @@ OUTPUT
       G
 
       bundle :install, :artifice => "endopint_marshal_fail_basic_authentication"
-      out.should_not include("#{user}:#{password}")
+      expect(out).not_to include("#{user}:#{password}")
       should_be_installed "rack 1.0.0"
     end
 
@@ -383,7 +441,74 @@ OUTPUT
       G
 
       bundle :install, :artifice => "endpoint_500"
-      out.should_not include("#{user}:#{password}")
+      expect(out).not_to include("#{user}:#{password}")
+    end
+
+    it "does not pass the user / password to different hosts on redirect" do
+      gemfile <<-G
+        source "#{basic_auth_source_uri}"
+        gem "rack"
+      G
+
+      bundle :install, :artifice => "endpoint_creds_diff_host"
+      should_be_installed "rack 1.0.0"
+    end
+
+    describe "with authentication details in bundle config" do
+      before do
+        gemfile <<-G
+          source "#{source_uri}"
+          gem "rack"
+        G
+      end
+
+      it "reads authentication details from bundle config" do
+        # The trailing slash is necessary here; Fetcher canonicalizes the URI.
+        bundle "config #{source_uri}/ #{user}:#{password}"
+
+        bundle :install, :artifice => "endpoint_strict_basic_authentication"
+
+        expect(out).to include("Fetching gem metadata from #{source_uri}")
+        should_be_installed "rack 1.0.0"
+      end
+
+      it "prefers auth supplied in the source uri" do
+        gemfile <<-G
+          source "#{basic_auth_source_uri}"
+          gem "rack"
+        G
+
+        bundle "config #{source_uri}/ otheruser:wrong"
+
+        bundle :install, :artifice => "endpoint_strict_basic_authentication"
+        should_be_installed "rack 1.0.0"
+      end
+
+      it "shows instructions if auth is not provided for the source" do
+        bundle :install, :artifice => "endpoint_strict_basic_authentication"
+        expect(out).to include("bundle config #{source_uri}/ username:password")
+      end
+
+      it "fails if authentication has already been provided, but failed" do
+        bundle "config #{source_uri}/ #{user}:wrong"
+
+        bundle :install, :artifice => "endpoint_strict_basic_authentication"
+        expect(out).to include("Bad username or password")
+      end
+    end
+
+    describe "with no password" do
+      let(:password) { nil }
+
+      it "passes basic authentication details" do
+        gemfile <<-G
+          source "#{basic_auth_source_uri}"
+          gem "rack"
+        G
+
+        bundle :install, :artifice => "endpoint_basic_authentication"
+        should_be_installed "rack 1.0.0"
+      end
     end
   end
 
@@ -395,15 +520,7 @@ OUTPUT
       bundled_app("broken_ssl").mkpath
       bundled_app("broken_ssl/openssl.rb").open("w") do |f|
         f.write <<-RUBY
-          $:.delete File.expand_path("..", __FILE__)
-          require 'openssl'
-
-          require 'bundler'
-          class Bundler::Fetcher
-            def fetch(*)
-              raise LoadError, "cannot load such file -- openssl"
-            end
-          end
+          raise LoadError, "cannot load such file -- openssl"
         RUBY
       end
     end
@@ -414,9 +531,51 @@ OUTPUT
         gem "rack"
       G
 
-      bundle :install, :artifice => "endpoint",
-        :env => {"RUBYOPT" => "-I#{bundled_app("broken_ssl")}"}
-      out.should include("Could not load OpenSSL.")
+      bundle :install, :env => {"RUBYOPT" => "-I#{bundled_app("broken_ssl")}"}
+      expect(out).to include("OpenSSL")
+    end
+  end
+
+  context "when SSL certificate verification fails" do
+    it "explains what happened" do
+      # Install a monkeypatch that reproduces the effects of openssl raising
+      # a certificate validation error when Rubygems tries to connect.
+      gemfile <<-G
+        class Net::HTTP
+          def start
+            raise OpenSSL::SSL::SSLError, "certificate verify failed"
+          end
+        end
+
+        source "#{source_uri.gsub(/http/, 'https')}"
+        gem "rack"
+      G
+
+      bundle :install
+      expect(out).to match(/could not verify the SSL certificate/i)
+    end
+  end
+
+  context ".gemrc with sources is present" do
+    before do
+      File.open(home('.gemrc'), 'w') do |file|
+        file.puts({:sources => ["https://rubygems.org"]}.to_yaml)
+      end
+    end
+
+    after do
+      home('.gemrc').rmtree
+    end
+
+    it "uses other sources declared in the Gemfile" do
+      gemfile <<-G
+        source "#{source_uri}"
+        gem 'rack'
+      G
+
+      bundle "install", :exitstatus => true, :artifice => "endpoint_marshal_fail"
+
+      expect(exitstatus).to eq(0)
     end
   end
 
