@@ -368,21 +368,18 @@ describe "Bundler.setup" do
     end
 
     it "does not randomly change the path when specifying --path and the bundle directory becomes read only" do
-      begin
-        bundle "install --path vendor/bundle"
+      bundle "install --path vendor/bundle"
 
-        Dir["**/*"].each do |f|
-          File.directory?(f) ?
-            File.chmod(0555, f) :
-            File.chmod(0444, f)
-        end
+      with_read_only("**/*") do
+        should_be_installed "rack 1.0.0"
+      end
+    end
+
+    it "finds git gem when default bundle path becomes read only" do
+      bundle "install"
+
+      with_read_only("#{Bundler.bundle_path}/**/*") do
         should_be_installed "rack 1.0.0"
-      ensure
-        Dir["**/*"].each do |f|
-          File.directory?(f) ?
-            File.chmod(0755, f) :
-            File.chmod(0644, f)
-        end
       end
     end
   end
